import numpy as np

from pynwb import register_class
from pynwb.core import NWBDataInterface, DynamicTable
from hdmf.utils import docval, getargs, popargs, get_docval


@register_class('Events', 'ndx-events')
class Events(NWBDataInterface):
    """
    A list of timestamps, stored in seconds, of an event.
    """

    __nwbfields__ = ('description',
                     'timestamps',
                     'resolution',
                     {'name': 'unit', 'settable': False})

    @docval({'name': 'name', 'type': str, 'doc': 'The name of this Events object'},  # required
            {'name': 'description', 'type': str, 'doc': 'The name of this Events object'},  # required
            {'name': 'timestamps', 'type': ('array_data', 'data'),  # required
             'doc': ('Event timestamps, in seconds, relative to the common experiment master-clock '
                     'stored in NWBFile.timestamps_reference_time.'),
             'shape': (None,)},
            {'name': 'resolution', 'type': float,
             'doc': ('The smallest possible difference between two event times. Usually 1 divided '
                     'by the event time sampling rate on the data acquisition system.'),
             'default': None})
    def __init__(self, **kwargs):
        description, timestamps, resolution = popargs('description', 'timestamps', 'resolution', kwargs)
        super().__init__(**kwargs)
        self.description = description
        self.timestamps = timestamps
        self.resolution = resolution
        self.fields['unit'] = 'seconds'


@register_class('LabeledEvents', 'ndx-events')
class LabeledEvents(Events):
    """
    A list of timestamps, stored in seconds, of an event that can have different
    labels. For example, this type could represent the times that reward was given,
    as well as which of three different types of reward was given. In this case, the
    'data' dataset would contain values {0, 1, 2}, and the 'labels' dataset
    would contain three text elements, where the first (index 0) specifies the name
    of the reward associated with data = 0, the second (index 1) specifies
    the name of the reward associated with data = 1, etc. The labels do not
    have to start at 0 and do not need to be sequential, e.g. the 'data' dataset
    could contain values {0, 10, 100}, and the 'labels' dataset could contain 101
    values, where labels[0] is 'No reward', labels[10] is '10% reward', labels[100]
    is 'Full reward', and all other entries in 'labels' are the empty string.
    """

    __nwbfields__ = ('data',
                     'labels')

    @docval(*get_docval(Events.__init__, 'name', 'description', 'timestamps'),  # required
            {'name': 'data', 'type': ('array_data', 'data'),  # required
             'doc': ("Unsigned integer labels that map onto strings using the mapping in the 'labels' dataset. "
                     "Values must be 0 or greater and need not be sequential. If a list/tuple/array of integer values "
                     "is passed, it will be converted to a numpy array of unsigned integer values. This dataset should "
                     "have the same number of elements as the 'timestamps' dataset."),
             'shape': (None,)},
            {'name': 'labels', 'type': ('array_data', 'data'),
             'doc': ("Mapping from an integer (the zero-based index) to a string, used to understand "
                     "the integer values in the 'data' dataset. Use an empty string to represent "
                     "a label value that is not mapped to any text. Use '' to represent any values "
                     "that are None or empty. If the argument is not specified, the label "
                     "will be set to the string representation of the data value and '' for other values."),
             'shape': (None,), 'default': None},
            *get_docval(Events.__init__, 'resolution'))
    def __init__(self, **kwargs):
        timestamps = getargs('timestamps', kwargs)
        data, labels = popargs('data', 'labels', kwargs)
        super().__init__(**kwargs)
        if len(timestamps) != len(data):
            raise ValueError('Timestamps and data must have the same length: %d != %d'
                             % (len(timestamps), len(data)))
        data = self.__check_label_indices_uint(data)
        self.data = data
        if labels is None:
            unique_indices = np.unique(data)
            self.labels = [''] * (max(unique_indices) + 1)
            for k in unique_indices:
                self.labels[k] = str(k)
        else:
            if None in labels:
                raise ValueError("None values are not allowed in the labels array. Please use '' for undefined labels.")
            self.labels = labels

    def __check_label_indices_uint(self, data):
        """Convert a list/tuple of integer label indices to a numpy array of unsigned integers. Raise error if negative
        or non-numeric values are found. If something other than a list/tuple/np.ndarray of ints or unsigned ints
        is provided, return the original array.
        """
        new_data = data
        if isinstance(new_data, (list, tuple)):
            new_data = np.array(new_data)
        if isinstance(new_data, np.ndarray):
            if not np.issubdtype(new_data.dtype, np.number):
                raise ValueError("'data' must be an array of numeric values that have type unsigned int or "
                                 "can be converted to unsigned int, not type %s" % new_data.dtype)
            if np.issubdtype(new_data.dtype, np.unsignedinteger):
                return new_data
            if (new_data < 0).any():
                raise ValueError("Negative values are not allowed in 'data'.")
            if np.issubdtype(new_data.dtype, np.integer):
                return new_data.astype(np.uint)
            # all other array dtypes will not be handled. the objectmapper will attempt to convert the data
        return data


@register_class('TTLs', 'ndx-events')
class TTLs(LabeledEvents):
    """
    Data type to hold timestamps of TTL pulses. The 'data' dataset contains the integer pulse values
    (or channel IDs), and the 'labels' dataset contains user-defined labels associated with each pulse
    value (or channel ID). The value at index i of the 'labels' dataset corresponds to a pulse value (or
    channel ID) of i in the 'data' dataset. For example, the first value (index 0) of the 'labels' dataset
    corresponds to a pulse value of 0. See the LabeledEvents type for more details.
    """
    pass


@register_class('AnnotatedEventsTable', 'ndx-events')
class AnnotatedEventsTable(DynamicTable):
    """
    Table to hold event timestamps and event metadata relevant to data preprocessing
    and analysis. Each row corresponds to a different event type. Use the 'event_time'
    dataset to store timestamps for each event type. Add user-defined columns to add
    metadata for each event type or event time.
    """

    __fields__ = (
        'resolution',
    )

    __columns__ = (
        {'name': 'event_times', 'description': 'Event times for each event type.', 'index': True},
        {'name': 'label', 'description': 'Label for each event type.'},
        {'name': 'event_description', 'description': 'Description for each event type.'}
        # note that the name 'description' cannot be used because it is already an attribute on VectorData
    )

    @docval({'name': 'description', 'type': str, 'doc': 'Description of what is in this table'},
            {'name': 'name', 'type': str, 'doc': 'Name of this AnnotatedEventsTable',
             'default': 'AnnotatedEventsTable'},
            {'name': 'resolution', 'type': float,
             'doc': ('The smallest possible difference between two event times. Usually 1 divided '
                     'by the event time sampling rate on the data acquisition system.'),
             'default': None},
            *get_docval(DynamicTable.__init__, 'id', 'columns', 'colnames'))
    def __init__(self, **kwargs):
        resolution = popargs('resolution', kwargs)
        super().__init__(**kwargs)
        self.resolution = resolution

    @docval({'name': 'label', 'type': str, 'doc': 'Label for each event type.'},
            {'name': 'event_description', 'type': str, 'doc': 'Description for each event type.'},
            {'name': 'event_times', 'type': 'array_data', 'doc': 'Event times for each event type.', 'shape': (None,)},
            {'name': 'id', 'type': int, 'doc': 'ID for each unit', 'default': None},
            allow_extra=True)
    def add_event_type(self, **kwargs):
        """Add an event type as a row to this table."""
        super().add_row(**kwargs)



def test_example_usage():
    from datetime import datetime

    from pynwb import NWBFile, NWBHDF5IO
    from ndx_events import LabeledEvents, AnnotatedEventsTable

    nwb = NWBFile(
        session_description='session description',
        identifier='cool_experiment_001',
        session_start_time=datetime.now().astimezone()
    )

    # create a new LabeledEvents type to hold events recorded from the data acquisition system
    events = LabeledEvents(
        name='LabeledEvents',
        description='events from my experiment',
        timestamps=[0., 0.5, 0.6, 2., 2.05, 3., 3.5, 3.6, 4.],
        resolution=1e-5,  # resolution of the timestamps, i.e., smallest possible difference between timestamps
        data=[0, 1, 2, 3, 5, 0, 1, 2, 4],
        labels=['trial_start', 'cue_onset', 'cue_offset', 'response_left', 'response_right', 'reward']
    )

    # add the LabeledEvents type to the acquisition group of the NWB file
    nwb.add_acquisition(events)

    # create a new AnnotatedEventsTable type to hold annotated events
    annotated_events = AnnotatedEventsTable(
        name='AnnotatedEventsTable',
        description='annotated events from my experiment',
        resolution=1e-5  # resolution of the timestamps, i.e., smallest possible difference between timestamps
    )
    # add a custom indexed (ragged) column to represent whether each event time was a bad event
    annotated_events.add_column(
        name='bad_event',
        description='whether each event time should be excluded',
        index=True
    )
    # add an event type (row) to the AnnotatedEventsTable instance
    annotated_events.add_event_type(
        label='Reward',
        event_description='Times when the subject received juice reward.',
        event_times=[1., 2., 3.],
        bad_event=[False, False, True],
        id=3
    )

    # create a processing module in the NWB file to hold processed events data
    events_module = nwb.create_processing_module(
        name='events',
        description='processed event data'
    )

    # add the AnnotatedEventsTable instance to the processing module
    events_module.add(annotated_events)

    # write nwb file
    filename = 'test.nwb'
    with NWBHDF5IO(filename, 'w') as io:
        io.write(nwb)

    # read nwb file and check its contents
    with NWBHDF5IO(filename, 'r', load_namespaces=True) as io:
        nwb = io.read()
        print(nwb)



import numpy as np

from pynwb.testing import TestCase
from pynwb.core import VectorData, VectorIndex

from ndx_events import Events, LabeledEvents, TTLs, AnnotatedEventsTable


class TestEvents(TestCase):

    def test_init(self):
        events = Events(
            name='Events',
            description='events from my experiment',
            timestamps=[0., 1., 2.],
            resolution=1e-5
        )
        self.assertEqual(events.name, 'Events')
        self.assertEqual(events.description, 'events from my experiment')
        self.assertEqual(events.timestamps, [0., 1., 2.])
        self.assertEqual(events.resolution, 1e-5)
        self.assertEqual(events.unit, 'seconds')


class TestLabeledEvents(TestCase):

    def test_init(self):
        events = LabeledEvents(
            name='LabeledEvents',
            description='events from my experiment',
            timestamps=[0., 1., 2.],
            resolution=1e-5,
            data=np.uint([3, 4, 3]),
            labels=['', '', '', 'event1', 'event2']
        )
        self.assertEqual(events.name, 'LabeledEvents')
        self.assertEqual(events.description, 'events from my experiment')
        self.assertEqual(events.timestamps, [0., 1., 2.])
        self.assertEqual(events.resolution, 1e-5)
        self.assertEqual(events.unit, 'seconds')
        np.testing.assert_array_equal(events.data, np.uint([3, 4, 3])),
        self.assertEqual(events.labels, ['', '', '', 'event1', 'event2'])

    def test_mismatch_length(self):
        msg = 'Timestamps and data must have the same length: 3 != 4'
        with self.assertRaisesWith(ValueError, msg):
            LabeledEvents(
                name='LabeledEvents',
                description='events from my experiment',
                timestamps=[0., 1., 2.],
                resolution=1e-5,
                data=np.uint([3, 4, 3, 5]),
                labels=['', '', '', 'event1', 'event2', 'event3']
            )

    def test_default_labels(self):
        events = LabeledEvents(
            name='LabeledEvents',
            description='events from my experiment',
            timestamps=[0., 1., 2.],
            resolution=1e-5,
            data=np.uint([3, 4, 3]),
        )
        self.assertEqual(events.labels, ['', '', '', '3', '4'])

    def test_none_in_labels(self):
        msg = "None values are not allowed in the labels array. Please use '' for undefined labels."
        with self.assertRaisesWith(ValueError, msg):
            LabeledEvents(
                name='LabeledEvents',
                description='events from my experiment',
                timestamps=[0., 1., 2.],
                resolution=1e-5,
                data=np.uint([3, 4, 3]),
                labels=[None, None, None, 'event1', 'event2']
            )

    def test_data_negative(self):
        msg = "Negative values are not allowed in 'data'."
        with self.assertRaisesWith(ValueError, msg):
            LabeledEvents(
                name='LabeledEvents',
                description='events from my experiment',
                timestamps=[0., 1., 2.],
                resolution=1e-5,
                data=[1, -2, 3],
                labels=['', '', '', 'event1', 'event2']
            )

    def test_data_int_conversion(self):
        le = LabeledEvents(
            name='LabeledEvents',
            description='events from my experiment',
            timestamps=[0., 1., 2.],
            resolution=1e-5,
            data=[1, 2, 3],
            labels=['', '', '', 'event1', 'event2']
        )
        np.testing.assert_array_equal(le.data, np.array([1, 2, 3]))
        self.assertEqual(le.data.dtype, np.uint)

    def test_data_string(self):
        msg = ("'data' must be an array of numeric values that have type unsigned int or "
               "can be converted to unsigned int, not type <U1")
        with self.assertRaisesWith(ValueError, msg):
            LabeledEvents(
                name='LabeledEvents',
                description='events from my experiment',
                timestamps=[0., 1., 2.],
                resolution=1e-5,
                data=['1', '2', '3'],
                labels=['', '', '', 'event1', 'event2']
            )

    def test_data_pass_through(self):
        data = [1.0, 2.0, 3.0]
        le = LabeledEvents(
            name='LabeledEvents',
            description='events from my experiment',
            timestamps=[0., 1., 2.],
            resolution=1e-5,
            data=data,
            labels=['', '', '', 'event1', 'event2']
        )
        self.assertIs(le.data, data)


class TestTTLs(TestCase):

    def test_init(self):
        events = TTLs(
            name='TTLs',
            description='ttl pulses from my experiment',
            timestamps=[0., 1., 2.],
            resolution=1e-5,
            data=np.uint([3, 4, 3]),
            labels=['', '', '', 'event1', 'event2']
        )
        self.assertEqual(events.name, 'TTLs')
        self.assertEqual(events.description, 'ttl pulses from my experiment')
        self.assertEqual(events.timestamps, [0., 1., 2.])
        self.assertEqual(events.resolution, 1e-5)
        self.assertEqual(events.unit, 'seconds')
        np.testing.assert_array_equal(events.data, np.uint([3, 4, 3])),
        self.assertEqual(events.labels, ['', '', '', 'event1', 'event2'])


class TestAnnotatedEventsTable(TestCase):

    def test_init(self):
        events = AnnotatedEventsTable(
            name='AnnotatedEventsTable',
            description='annotated events from my experiment',
            resolution=1e-5
        )
        self.assertEqual(events.name, 'AnnotatedEventsTable')
        self.assertEqual(events.description, 'annotated events from my experiment')
        self.assertEqual(events.resolution, 1e-5)

    def test_add_event_type(self):
        events = AnnotatedEventsTable(
            name='AnnotatedEventsTable',
            description='annotated events from my experiment'
        )
        events.add_event_type(
            label='Reward',
            event_description='Times when the animal received juice reward.',
            event_times=[1., 2., 3.],
            id=3
        )
        events.add_event_type(
            label='Abort',
            event_description='Times when the animal aborted the trial.',
            event_times=[0.5, 4.5],
            id=4
        )
        self.assertEqual(events.id.data, [3, 4])
        self.assertEqual(events['event_times'][0], [1., 2., 3.])
        self.assertEqual(events['event_times'][1], [0.5, 4.5])
        self.assertEqual(events['label'][0], 'Reward')
        self.assertEqual(events['label'][1], 'Abort')
        self.assertListEqual(events['event_description'].data, ['Times when the animal received juice reward.',
                                                                'Times when the animal aborted the trial.'])
        self.assertEqual(events.colnames, ('event_times', 'label', 'event_description'))
        self.assertEqual(len(events.columns), 4)
        self.assertEqual(events.columns[0].name, 'event_times_index')
        self.assertIsInstance(events.columns[0], VectorIndex)
        self.assertIs(events.columns[0].target, events.columns[1])
        self.assertEqual(events.columns[1].name, 'event_times')
        self.assertIsInstance(events.columns[1], VectorData)
        self.assertEqual(events.columns[2].name, 'label')
        self.assertIsInstance(events.columns[2], VectorData)
        self.assertEqual(events.columns[3].name, 'event_description')
        self.assertIsInstance(events.columns[3], VectorData)
        self.assertEqual(events.resolution, None)



import datetime
import numpy as np
from pynwb import NWBFile, NWBHDF5IO
from pynwb.testing import AcquisitionH5IOMixin, TestCase, remove_test_file

from ndx_events import Events, LabeledEvents, TTLs, AnnotatedEventsTable


class TestEventsIOSimple(TestCase):
    """Simple roundtrip test for CSD."""

    def setUp(self):
        self.nwbfile = NWBFile(
            session_description='session_description',
            identifier='identifier',
            session_start_time=datetime.datetime.now(datetime.timezone.utc)
        )
        self.path = 'test.nwb'

    def tearDown(self):
        remove_test_file(self.path)

    def test_roundtrip(self):
        """
        Add a CSD to an "ecephys" processing module in the NWBFile, write it to file, read the file, and test that the
        CSD from the file matches the original CSD.
        """

        events = Events(
            name='Events',
            description='events from my experiment',
            timestamps=[0., 1., 2.],
            resolution=1e-5
        )
        self.nwbfile.add_acquisition(events)

        labeled_events = LabeledEvents(
            name='LabeledEvents',
            description='events from my experiment',
            timestamps=[0., 1., 2.],
            resolution=1e-5,
            data=np.uint([3, 4, 3]),
            labels=['', '', '', 'event1', 'event2']
        )
        self.nwbfile.add_acquisition(labeled_events)

        ttls = TTLs(
            name='TTLs',
            description='events from my experiment',
            timestamps=[0., 1., 2.],
            resolution=1e-5,
            data=np.uint([3, 4, 3]),
            labels=['', '', '', 'event1', 'event2']
        )
        self.nwbfile.add_acquisition(ttls)

        annotated_events = AnnotatedEventsTable(
            name='AnnotatedEventsTable',
            description='annotated events from my experiment',
            resolution=1e-5
        )
        annotated_events.add_column(
            name='extra',
            description='extra metadata for each event type'
        )
        annotated_events.add_event_type(
            label='Reward',
            event_description='Times when the animal received juice reward.',
            event_times=[1., 2., 3.],
            extra='extra',
            id=3
        )
        annotated_events.add_event_type(
            label='Nosepoke',
            event_description='Times when the animal poked its noise through the input port.',
            event_times=[1., 2., 3.],
            extra='extra',
            id=5
        )

        events_module = self.nwbfile.create_processing_module(
            name='events',
            description='processed events data'
        )
        events_module.add(annotated_events)

        with NWBHDF5IO(self.path, mode='w') as io:
            io.write(self.nwbfile)

        with NWBHDF5IO(self.path, mode='r', load_namespaces=True) as io:
            read_nwbfile = io.read()
            self.assertContainerEqual(events, read_nwbfile.acquisition['Events'])
            self.assertContainerEqual(labeled_events, read_nwbfile.acquisition['LabeledEvents'])
            self.assertContainerEqual(ttls, read_nwbfile.acquisition['TTLs'])
            self.assertContainerEqual(annotated_events, read_nwbfile.processing['events']['AnnotatedEventsTable'])


class TestEventsIO(AcquisitionH5IOMixin, TestCase):

    def setUpContainer(self):
        """ Return the test Events to read/write """
        events = Events(
            name='Events',
            description='events from my experiment',
            timestamps=[0., 1., 2.],
            resolution=1e-5
        )
        return events


class TestLabeledEventsIO(AcquisitionH5IOMixin, TestCase):

    def setUpContainer(self):
        """ Return the test Events to read/write """
        labeled_events = LabeledEvents(
            name='LabeledEvents',
            description='events from my experiment',
            timestamps=[0., 1., 2.],
            resolution=1e-5,
            data=np.uint([3, 4, 3]),
            labels=['', '', '', 'event1', 'event2']
        )
        return labeled_events


class TestTTLs(AcquisitionH5IOMixin, TestCase):

    def setUpContainer(self):
        """ Return the test Events to read/write """
        ttls = TTLs(
            name='TTLs',
            description='events from my experiment',
            timestamps=[0., 1., 2.],
            resolution=1e-5,
            data=np.uint([3, 4, 3]),
            labels=['', '', '', 'event1', 'event2']
        )
        return ttls


class TestAnnotatedEventsTableIO(AcquisitionH5IOMixin, TestCase):
    """ Test adding AnnotatedEventsTable into acquisition and accessing AnnotatedEvents after read """

    def setUpContainer(self):
        """ Return the test AnnotatedEventsTable to read/write """
        annotated_events = AnnotatedEventsTable(
            name='AnnotatedEventsTable',
            description='annotated events from my experiment',
            resolution=1e-5
        )
        annotated_events.add_column(
            name='extra',
            description='extra metadata for each event type'
        )
        annotated_events.add_event_type(
            label='Reward',
            event_description='Times when the animal received juice reward.',
            event_times=[1., 2., 3.],
            extra='extra',
            id=3
        )
        annotated_events.add_event_type(
            label='Nosepoke',
            event_description='Times when the animal poked its noise through the input port.',
            event_times=[1., 2., 3.],
            extra='extra',
            id=5
        )
        return annotated_events



from pynwb import register_map
from pynwb.io.core import NWBContainerMapper, VectorDataMap as PynwbVectorDataMap
from hdmf.common.io.table import DynamicTableMap
from hdmf.build import BuildManager
from hdmf.common import VectorData
from hdmf.utils import getargs, docval
from hdmf.spec import AttributeSpec

from ..events import Events, LabeledEvents, AnnotatedEventsTable


@register_map(Events)
class EventsMap(NWBContainerMapper):

    def __init__(self, spec):
        super().__init__(spec)
        # map object attribute Events.unit -> spec Events/timestamps.unit
        # map object attribute Events.resolution -> spec Events/timestamps.resolution
        timestamps_spec = self.spec.get_dataset('timestamps')
        self.map_spec('unit', timestamps_spec.get_attribute('unit'))
        self.map_spec('resolution', timestamps_spec.get_attribute('resolution'))


@register_map(LabeledEvents)
class LabeledEventsMap(EventsMap):

    def __init__(self, spec):
        super().__init__(spec)
        # map object attribute LabeledEvents.labels -> spec LabeledEvents/data.labels
        data_spec = self.spec.get_dataset('data')
        self.map_spec('labels', data_spec.get_attribute('labels'))


@register_map(AnnotatedEventsTable)
class AnnotatedEventsTableMap(DynamicTableMap):

    def __init__(self, spec):
        super().__init__(spec)
        # map object attribute AnnotatedEventsTable.resolution -> spec AnnotatedEventsTable/event_times.resolution
        event_times_spec = self.spec.get_dataset('event_times')
        self.map_spec('resolution', event_times_spec.get_attribute('resolution'))

    @DynamicTableMap.constructor_arg('resolution')
    def resolution_carg(self, builder, manager):
        # on construct, map builder for AnnotatedEventsTable.datasets['event_times'].attributes['resolution']
        # -> AnnotatedEventsTable.__init__ argument 'resolution'
        if 'event_times' in builder:
            return builder['event_times'].attributes.get('resolution')
        return None


@register_map(VectorData)
class VectorDataMap(PynwbVectorDataMap):

    # TODO when merging into NWB core, fold this into pynwb.io.core.VectorDataMap

    @docval({"name": "spec", "type": AttributeSpec, "doc": "the spec to get the attribute value for"},
            {"name": "container", "type": VectorData, "doc": "the container to get the attribute value from"},
            {"name": "manager", "type": BuildManager, "doc": "the BuildManager used for managing this build"},
            returns='the value of the attribute')
    def get_attr_value(self, **kwargs):
        ''' Get the value of the attribute corresponding to this spec from the given container '''
        spec, container, manager = getargs('spec', 'container', 'manager', kwargs)

        # on build of VectorData objects, map object attribute AnnotatedEventsTable.resolution
        # -> spec AnnotatedEventsTable/event_times.resolution
        if isinstance(container.parent, AnnotatedEventsTable):
            if container.name == 'event_times':
                if spec.name == 'resolution':
                    return container.parent.resolution
        return super().get_attr_value(**kwargs)
